import os
from typing import Dict, Optional

from pydantic import BaseSettings


class Settings(BaseSettings):
    API_V1_STR: str = "/api/v1"
    LOCAL_STACK: bool = False
    E2E: bool = False
    CI_RUNNER_NAME: Optional[str] = None

    VERBOSE: bool = False
    DEBUG: bool = False
    WEB_URL: str = "https://app.taktile.com"
    TAKTILE_API_URL: str = "https://taktile-api.taktile.com"
    DEPLOYMENT_API_URL: str = "https://deployment-api.taktile.com"

    TKTL_CONFIG_PATH: str = os.path.expanduser("~/.config/tktl")
    CONFIG_FILE_NAME: str = "config.json"
    TAKTILE_API_KEY: Optional[str] = None
    HELP_HEADERS_COLOR: str = "yellow"
    HELP_OPTIONS_COLOR: str = "green"
    USE_CONSOLE_COLORS: bool = True
    HELP_COLORS_DICT: Dict = {
        "help_headers_color": "yellow",
        "help_options_color": "green",
    }

    LOCAL_ARROW_ENDPOINT: str = "grpc+tcp://127.0.0.1:5005"
    LOCAL_REST_ENDPOINT: str = "http://127.0.0.1:8080"
    ARROW_BATCH_MB: int = 50
    PARQUET_BATCH_DEFAULT_ROWS_READ: int = int(1e6)

    class Config:
        case_sensitive = True


settings = Settings()



import os

import pytest

from tktl.login import login, logout


@pytest.fixture
def user_key():
    return os.environ["TEST_USER"], os.environ["TEST_USER_API_KEY"]


@pytest.fixture
def test_user_deployed_repos():
    return (
        ("tktl-admin/other-project", "testbranch", "repayment"),
        ("tktl-admin/other-project", "master", "repayment"),
        ("tktl-admin/sample-project", "master", "repayment"),
        ("tktl-admin/integ-testing", "master", "repayment"),
        ("tktl-admin/grpc-test", "main", "repayment"),
    )


@pytest.fixture(scope="function")
def logged_in_context():
    yield login(os.environ["TEST_USER_API_KEY"])
    logout()


@pytest.fixture()
def sample_deployed_url():
    return "tktl-adm-grpc-test-main-1a485d0d54.stg-saas.taktile.com"



import click
import colorama  # type: ignore
from click._compat import get_text_stderr  # type: ignore

import tktl.cli.login  # noqa: F401


def show(self, file=None):
    if file is None:
        file = get_text_stderr()
    color = None
    hint = ""
    if self.cmd is not None and self.cmd.get_help_option(self.ctx) is not None:
        hint = 'Try "%s %s" for help.\n' % (
            self.ctx.command_path,
            self.ctx.help_option_names[0],
        )
    if self.ctx is not None:
        color = self.ctx.color
        click.echo(self.ctx.get_usage() + "\n%s" % hint, file=file, color=color)
    msg = (
        colorama.Fore.RED
        + "Error: %s" % self.format_message()
        + colorama.Style.RESET_ALL
    )
    click.echo(msg, file=file, color=color)


# not-very-elegant way to color click's error messages
click.exceptions.UsageError.show = show  # type: ignore



import random

import numpy
import pandas
import pandas.api.types as ptypes
import pytest
from pydantic import ValidationError

from tktl.core.serializers.base import CustomDeserializingModelT
from tktl.core.serializers.rest.typed import replace_nans_and_infs, to_pydantic
from tktl.core.t import RestSchemaTypes


def random_string():
    return "".join([random.choice("abcde") for _ in range(10)])


def test_from_single_value():
    id_ = random_string()
    schema = to_pydantic(342314, unique_id=id_)
    assert schema.__name__ == f"SingleValue__{id_}"
    assert schema.validate({"value": 1})
    deser = schema(**{"value": 1}).deserialize()
    assert deser == 1


def test_from_series(serializer_df_inputs: pandas.DataFrame):
    series = serializer_df_inputs.A
    id_ = random_string()
    schema = to_pydantic(series, unique_id=id_)
    assert schema.__name__ == f"Series__{id_}"
    rand_values = {"A": [1, 1, 12, 1, 1, 321, None]}
    assert schema.validate(rand_values)
    deser = schema(**rand_values).deserialize()
    assert isinstance(deser, pandas.Series)
    assert ptypes.is_float_dtype(deser.dtype)


def test_from_frame(serializer_df_inputs: pandas.DataFrame):
    id_ = random_string()
    schema = to_pydantic(serializer_df_inputs, unique_id=id_)
    assert schema.__name__ == f"DataFrame__{id_}"
    rand_values = [
        {k: numpy.random.randn(1)[0] for k in serializer_df_inputs.columns}
    ] * 10
    assert schema.validate(rand_values)
    deser = schema.parse_obj(rand_values).deserialize()
    assert isinstance(deser, pandas.DataFrame)
    for col in deser.columns:
        assert ptypes.is_float_dtype(deser[col].dtype)


def test_from_array_1d():
    arr = numpy.random.randn(100)
    id_ = random_string()
    schema = to_pydantic(arr, unique_id=id_)
    assert schema.__name__ == f"{RestSchemaTypes.FLAT_ARRAY.value}__{id_}"
    assert isinstance(schema(value=[1]).__root__, list)
    assert isinstance(schema(value=[1, 10000.0]).__root__, list)


def test_from_multi_dim_array():
    arr = numpy.random.randint(0, 100, size=(10, 5, 2, 2))
    schema = to_pydantic(arr)
    assert schema.__name__ == RestSchemaTypes.ARRAY.value
    sample = numpy.random.randint(0, 100, size=(5, 2, 2)).tolist()
    assert schema.validate([sample])
    with pytest.raises(ValidationError):
        assert schema.validate([[[312, 432], [312, 432]]])

    deser = schema.parse_obj([sample]).deserialize()
    assert isinstance(deser, numpy.ndarray)


def test_nan_handling():
    arr = numpy.random.randn(100)
    arr[0:5] = numpy.nan
    arr[5:10] = numpy.NaN
    schema = to_pydantic(arr)
    assert schema.__name__ == RestSchemaTypes.FLAT_ARRAY.value
    assert isinstance(schema(value=[1]).__root__, list)
    assert schema.schema()["properties"]["value"]["items"]["type"] == "number"
    assert schema.schema()["properties"]["value"]["items"]["type"] == "number"


def test_replace():
    arr = numpy.array([[numpy.nan, 1.1, 1.1], [numpy.inf, 1.1, 3.1]])
    assert replace_nans_and_infs(arr).tolist() == [[None, 1.1, 1.1], [0.0, 1.1, 3.1]]


def test_nan_handling_multi_dim():
    arr = numpy.array(
        [[[numpy.nan, 1.1, numpy.inf], [12.1, 1, 3]], [[1, numpy.nan, 1], [None, 1, 3]]]
    )
    schema = to_pydantic(arr)
    assert schema.__name__ == RestSchemaTypes.ARRAY.value
    sample = numpy.random.randn(2, 3).tolist()
    assert schema.validate([sample])
    print(schema.schema())
    assert schema.schema()["items"]["items"]["items"]["type"] == "number"
    assert schema.schema()["example"] == [[[None, 1.1, 0.0], [12.1, 1, 3]]]
    with pytest.raises(ValidationError):
        assert schema.validate([[[312, 432], [312, 432]]])

    deser = schema.parse_obj([sample]).deserialize()
    assert isinstance(deser, numpy.ndarray)


def test_from_sequence():
    sequence = numpy.random.randint(0, 5, size=(10, 2, 2)).tolist()
    schema = to_pydantic(sequence)
    with pytest.raises(ValidationError):
        schema.parse_obj([[1, 2], [3]])
    with pytest.raises(ValidationError):
        schema.parse_obj([[3, 1, 3]])

    parsed = schema.parse_obj([[1, 2], [3, 4]])
    assert isinstance(parsed, CustomDeserializingModelT)
    assert parsed.deserialize() == [[1, 2], [3, 4]]

    deeply_nested_sequence = numpy.random.randint(0, 5, size=(10, 2, 2, 2, 2)).tolist()
    schema = to_pydantic(deeply_nested_sequence)
    parsed = schema.parse_obj(
        [[[[1, 2], [3, 4]], [[1, 2], [3, 4]]], [[[1, 2], [3, 4]], [[1, 2], [3, 4]]]]
    )
    assert parsed.deserialize() == [
        [[[1, 2], [3, 4]], [[1, 2], [3, 4]]],
        [[[1, 2], [3, 4]], [[1, 2], [3, 4]]],
    ]


def test_from_dict_or_list_of_dicts():
    sequence = [{"a": 1, "b": 2}, {"a": 1, "b": 100}]
    mapping_model = to_pydantic(sequence)
    valid = mapping_model(**{"a": 10, "b": 20})
    assert valid.deserialize() == {"a": 10, "b": 20}

    with pytest.raises(ValidationError):
        print(mapping_model(**{"c": 20, "b": "sdfgasdfg"}))



from click.testing import CliRunner

from tktl import main
from tktl.cli import validate


def test_validate():
    """Test the validation cli."""
    runner = CliRunner()
    result = runner.invoke(main.validate)
    assert result.exit_code == 0
    assert "help" in result.output
    assert "config" in result.output
    assert "all" in result.output
    assert "import" in result.output
    assert "unittest" in result.output
    assert "integration" in result.output
    assert "profiling" in result.output


def test_validate_config():
    """Test the validation cli."""
    runner = CliRunner()
    result = runner.invoke(validate.validate_config_command, ["."])
    assert result.exit_code == 2


def test_validate_all():
    """Test the validation cli."""
    runner = CliRunner()
    result = runner.invoke(validate.validate_all_command, ["."])
    assert result.exit_code == 2


def test_validate_import():
    """Test the validation cli."""
    runner = CliRunner()
    result = runner.invoke(validate.validate_import_command, ["."])
    assert result.exit_code == 2


def test_validate_unittest():
    """Test the validation cli."""
    runner = CliRunner()
    result = runner.invoke(validate.validate_unittest_command, ["."])
    assert result.exit_code == 2


def test_validate_integration():
    """Test the validation cli."""
    runner = CliRunner()
    result = runner.invoke(validate.validate_integration_command, ["."])
    assert result.exit_code == 2


def test_validate_profiling():
    """Test the validation cli."""
    runner = CliRunner()
    result = runner.invoke(validate.validate_profiling_command, ["."])
    assert result.exit_code == 2



import functools
import json
from io import BytesIO
from typing import Dict, List, Sequence, Type, Union

import pyarrow  # type: ignore
from pyarrow import json as j


def _serialize_dicts_or_list_of_dicts(
    values: Union[Sequence[Dict], Dict]
) -> pyarrow.Table:
    mem_file = BytesIO()
    if isinstance(values, dict):
        values = [values]
    for item in values:
        mem_file.write(json.dumps(item).encode())
    mem_file.seek(0)
    return j.read_json(mem_file)


def _get_final_serializer(initial_type: Type, final_type: Type):
    if initial_type != final_type:
        if initial_type == list:
            return "NATIVE_list"
        elif initial_type == tuple:
            return "FINAL_tuple"
        elif initial_type == dict:
            return "FINAL_first"
        else:
            raise ValueError()
    return None


def add_deserializing_metadata(
    table: pyarrow.Table,
    *,
    deserializers: List[str] = None,
    serializing_cls: str = None
) -> pyarrow.Table:
    metadata: Dict[str, Union[str, bytes]] = {}
    if deserializers:
        metadata["DESERIALIZERS"] = json.dumps(deserializers).encode()
    if serializing_cls:
        metadata["CLS"] = serializing_cls
    return _replace_table_with_metadata(table=table, metadata=metadata)


def _replace_table_with_metadata(table: pyarrow.Table, metadata: Dict):
    final = {**table.schema.metadata, **metadata} if table.schema.metadata else metadata
    return table.replace_schema_metadata(metadata=final)


def inject_metadata(serializer):
    @functools.wraps(serializer)
    def wraps(cls, *args, **kwargs):
        table: pyarrow.Table = serializer(cls, *args, **kwargs)
        table = add_deserializing_metadata(table=table, serializing_cls=cls.__name__)
        return table

    return wraps



import os

import pandas
import pytest
from pandas.testing import assert_index_equal

from tktl.core.clients.arrow import ArrowFlightClient
from tktl.core.exceptions import TaktileSdkError
from tktl.core.managers.auth import AuthConfigManager


def test_instantiate_client():

    key = os.environ["TEST_USER_API_KEY"]
    AuthConfigManager.set_api_key(key)

    with pytest.raises(TaktileSdkError):
        ArrowFlightClient(
            api_key=key,
            repository_name=f"{os.environ['TEST_USER']}/test-new",
            branch_name="master",
            endpoint_name="repayment",
        )
    client = ArrowFlightClient(
        api_key=key,
        repository_name=f"{os.environ['TEST_USER']}/integ-testing",
        branch_name="master",
        endpoint_name="repayment",
    )
    assert client.location is not None


def test_instantiate_by_url(sample_deployed_url):
    key = os.environ["TEST_USER_API_KEY"]
    AuthConfigManager.set_api_key(key)
    client = ArrowFlightClient.for_url(
        api_key=key, url=sample_deployed_url, endpoint_name="repayment"
    )
    assert client.location == f"grpc+tls://{sample_deployed_url}:5005"


def test_predict_flow(sample_deployed_url, capsys):
    key = os.environ["TEST_USER_API_KEY"]
    AuthConfigManager.set_api_key(key)
    client = ArrowFlightClient(
        api_key=key,
        repository_name="tktl-admin/grpc-test",
        endpoint_name="repayment_input_idx",
        branch_name="main",
    )
    x, y = client.get_sample_data()
    small = x.head()
    assert isinstance(x, pandas.DataFrame)
    predictions = client.predict(small.head(10))
    assert isinstance(predictions, pandas.Series)

    small.index = pandas.Series(["A"] * len(small))
    predictions = client.predict(small, use_input_index=True)
    assert_index_equal(predictions.index, small.index)

    client = ArrowFlightClient(
        api_key=key,
        repository_name="tktl-admin/grpc-test",
        endpoint_name="repayment",
        branch_name="main",
    )
    client.predict(small.head(10), use_input_index=True)
    out, err = capsys.readouterr()
    assert (
        "Inputs or Outputs are not of type series or dataframe, use_input_index has no effect\n"
        in out
    )



import logging

import click

from tktl.commands.init import init_project

_logger = logging.getLogger("root")


@click.command(help="Creates project scaffolding")
@click.option(
    "--name", help="Name of the project", default="tktl-serving", required=True
)
@click.option(
    "--path", help="Directory where the new project will be created", required=False
)
def init(path: str, name: str) -> None:
    """Creates a new project with the necessary scaffolding.

    Creates the supporting files needed. The directory structure of a new
    project , and the files within it will look like this:

        .dockerignore
        .gitattributes
        .gitignore
        .buildfile
        README.md
        assets              # Where your ML models and test data live
        requirements.txt    # User-specified requirements
        src                 # Source code for endpoint definitions
        tests               # User-specified tests
        tktl.yaml           # Taktile configuration options

    """
    init_project(path=path, name=name)



from tktl.core.clients import DeploymentApiClient
from tktl.core.clients.arrow import ArrowFlightClient
from tktl.core.clients.rest import RestClient
from tktl.core.clients.taktile import TaktileApiClient
from tktl.core.loggers import LOG, Logger
from tktl.core.managers.auth import AuthConfigManager
from tktl.core.t import ServiceT


class CommandBase(object):
    def __init__(self, api=None):
        self.api = api


class BaseClientCommand:
    def __init__(
        self,
        kind: ServiceT,
        api_key: str,
        repository: str,
        branch_name: str,
        endpoint_name: str = "",
        local: bool = False,
        logger: Logger = LOG,
        skip_auth: bool = False,
    ):
        self.kind = kind
        self.client_params = {
            "api_key": api_key,
            "repository_name": repository,
            "branch_name": branch_name,
            "endpoint_name": endpoint_name,
            "local": local,
            "logger": logger,
            "skip_auth": skip_auth,
        }
        self.client = self._get_client()

    def execute(self, *args, **kwargs):
        pass

    def _get_client(self):
        return (
            RestClient(**self.client_params)
            if self.kind == ServiceT.REST
            else ArrowFlightClient(**self.client_params)
        )


class BaseDeploymentApiCommand:
    def __init__(self, logger: Logger = LOG):
        self.client = DeploymentApiClient(logger=logger)


class BaseTaktileApiCommand:
    def __init__(self, logger: Logger = LOG):
        self.client = TaktileApiClient(logger=logger)


class BaseRestClientCommand(BaseClientCommand):
    def __init__(
        self,
        repository: str,
        branch_name: str,
        local: bool,
        logger: Logger = LOG,
        skip_auth: bool = False,
    ):
        super().__init__(
            kind=ServiceT.REST,
            api_key=AuthConfigManager.get_api_key(),
            repository=repository,
            branch_name=branch_name,
            local=local,
            logger=logger,
            skip_auth=skip_auth,
        )


class BaseGrpcClientCommand(BaseClientCommand):
    def __init__(
        self,
        repository: str,
        branch_name: str,
        local: bool,
        logger: Logger = LOG,
        skip_auth: bool = False,
    ):
        super().__init__(
            kind=ServiceT.GRPC,
            api_key=AuthConfigManager.get_api_key(),
            repository=repository,
            branch_name=branch_name,
            local=local,
            logger=logger,
            skip_auth=skip_auth,
        )



class ApplicationError(Exception):
    ...


class TaktileSdkError(Exception):
    ...


class AuthenticationError(ApplicationError):
    ...


class ArchiveUploadError(TaktileSdkError):
    ...


class PresignedUrlAccessDeniedError(ArchiveUploadError):
    ...


class ResourceFetchingError(TaktileSdkError):
    ...


class ResourceCreatingError(TaktileSdkError):
    ...


class MalformedResponseError(TaktileSdkError):
    ...


class ResourceCreatingDataError(ResourceCreatingError):
    ...


class PresignedUrlMalformedResponseError(ArchiveUploadError):
    ...


class PresignedUrlError(ArchiveUploadError):
    ...


class ProjectAccessDeniedError(ArchiveUploadError):
    ...


class WrongPathError(ArchiveUploadError):
    ...


class PresignedUrlUnreachableError(ArchiveUploadError):
    ...


class PresignedUrlConnectionError(ArchiveUploadError):
    ...


class InvalidParametersError(TaktileSdkError):
    ...


class BadResponseError(ApplicationError):
    ...


class InvalidInputError(ApplicationError):
    ...


class MutuallyExclusiveParametersUsedError(Exception):
    ...


class SerializerException(Exception):
    ...


class UnsupportedInputTypeException(Exception):
    ...
